from thumbor.filters import BaseFilter, filter_method, PHASE_PRE_LOAD
from thumbor.utils import logger


ALLOWED_FORMATS = [
    'png', 'jpeg', 'jpg', 'gif', 'webp', 'webm', 'mp4', 'hevc', 'h264', 'h265', 'vp9']


class Filter(BaseFilter):
    phase = PHASE_PRE_LOAD

    @filter_method(BaseFilter.String)
    def format(self, format):
        logger.warn('Setting format to %s' % format)
        if format.lower() not in ALLOWED_FORMATS:
            logger.debug('Format not allowed: %s' % format.lower())
            self.context.request.format = None
        else:
            logger.debug('Format specified: %s' % format.lower())
            self.context.request.format = format.lower()



from thumbor.filters import BaseFilter, filter_method, PHASE_PRE_LOAD


class Filter(BaseFilter):
    phase = PHASE_PRE_LOAD

    @filter_method(BaseFilter.String)
    def tune(self, value):
        self.context.request.tune = value



from thumbor.filters import BaseFilter, filter_method, PHASE_PRE_LOAD
from thumbor.utils import logger


class Filter(BaseFilter):
    phase = PHASE_PRE_LOAD

    @filter_method(r'(?P<pos>\-?(?:(?:\d\d:)?\d\d:\d\d(?:\.\d+?)?|\d+(?:\.\d+?)?))')
    def still(self, position='0'):
        logger.debug('Setting still frame at position %s' % position)
        self.context.request.still_position = position



from contextlib import contextmanager
import os
import shutil
from struct import unpack
from tempfile import NamedTemporaryFile, mkdtemp


@contextmanager
def named_tmp_file(data=None, extension=None, **kwargs):
    kwargs.setdefault('delete', False)
    if extension is not None:
        kwargs.setdefault('suffix', extension)
    f = NamedTemporaryFile(**kwargs)
    try:
        if data:
            f.write(data)
        f.close()
        yield f.name
    finally:
        os.unlink(f.name)


@contextmanager
def make_tmp_dir():
    try:
        tmp_dir = mkdtemp()
        yield tmp_dir
    finally:
        shutil.rmtree(tmp_dir, ignore_errors=True)


def is_mp4(buf):
    if buf[4:8] != b'ftyp':
        return False
    (ftyp_box_len,) = unpack('>L', buf[:4])
    if not (20 <= ftyp_box_len <= 256) or (ftyp_box_len % 4) != 0:
        return False
    major_brand = unpack('4s', buf[8:12])
    compat_brand_len = ftyp_box_len - 16
    fmt = '4s' * (compat_brand_len // 4)
    compat_brands = unpack(fmt, buf[16:ftyp_box_len])
    all_brands = set(major_brand + compat_brands)
    return bool(all_brands & {b'isom', b'avc1', b'iso2', b'mp41', b'mp42'})



from io import BytesIO

from PIL import Image
from thumbor.engines import BaseEngine
from thumbor.utils import logger

from thumbor_video_engine.utils import named_tmp_file, is_mp4


def patch_baseengine_get_mimetype():
    """
    Monkey-patch BaseEngine.get_mimetype() to recognize all mp4 files as video/mp4
    """
    orig_get_mimetype = BaseEngine.get_mimetype

    def get_mimetype(cls, buffer):
        mimetype = orig_get_mimetype(buffer)
        if mimetype is not None:
            return mimetype
        elif is_mp4(buffer):
            return 'video/mp4'

    BaseEngine.get_mimetype = classmethod(get_mimetype)


patch_baseengine_get_mimetype()


def is_animated(buffer):
    im = Image.open(BytesIO(buffer))
    return getattr(im, 'is_animated', False)


class Engine(object):
    """An engine that dispatches video files to ffmpeg, and others to PIL"""

    def __init__(self, context):
        self.engine = None
        self.context = context
        self.ffmpeg_handle_animated_gif = context.config.FFMPEG_HANDLE_ANIMATED_GIF
        self.ffmpeg_handle_animated_webp = True
        self.use_gif_engine = context.config.FFMPEG_USE_GIFSICLE_ENGINE

    @property
    def image_engine(self):
        if not hasattr(self.context.modules, 'image_engine'):
            # Instantiate the image engine class from the config (default is
            # thumbor.engines.pil)
            self.context.modules.importer.import_item('IMAGE_ENGINE', 'Engine')
            self.context.modules.image_engine = (
                self.context.modules.importer.image_engine(self.context))
        return self.context.modules.image_engine

    @property
    def ffmpeg_engine(self):
        if not hasattr(self.context.modules, 'ffmpeg_engine'):
            # Instantiate the video engine class from the config (default is
            # thumbor_video_engine.engines.ffmpeg)
            self.context.modules.importer.import_item('FFMPEG_ENGINE', 'Engine')
            self.context.modules.ffmpeg_engine = (
                self.context.modules.importer.ffmpeg_engine(self.context))
        return self.context.modules.ffmpeg_engine

    def load(self, buffer, extension):
        mime = BaseEngine.get_mimetype(buffer)

        is_gif = extension == '.gif'
        is_webp = extension == '.webp'

        if is_webp and self.ffmpeg_handle_animated_webp and is_animated(buffer):
            logger.debug("Setting engine to %s (extension %s)" % (
                self.context.config.FFMPEG_ENGINE, extension))
            self.engine = self.ffmpeg_engine
        elif is_gif and self.ffmpeg_handle_animated_gif and is_animated(buffer):
            logger.debug("Setting engine to %s (extension %s)" % (
                self.context.config.FFMPEG_ENGINE, extension))
            self.engine = self.ffmpeg_engine
        elif is_gif and self.use_gif_engine:
            logger.debug("Setting engine to %s (extension %s)" % (
                self.context.config.GIF_ENGINE, extension))
            self.engine = self.context.modules.gif_engine
        elif mime.startswith('video/'):
            logger.debug("Setting engine to %s (extension %s)" % (
                self.context.config.FFMPEG_ENGINE, extension))
            self.engine = self.ffmpeg_engine
        else:
            logger.debug("Setting engine to %s (extension %s)" % (
                self.context.config.IMAGE_ENGINE, extension))
            self.engine = self.image_engine

        still_frame_pos = getattr(self.context.request, 'still_position', None)
        # Are we requesting a still frame?
        if self.engine is self.ffmpeg_engine and still_frame_pos:
            with named_tmp_file(data=buffer, suffix=extension) as src_file:
                buffer = self.ffmpeg_engine.run_ffmpeg(
                    src_file, 'png', ['-ss', still_frame_pos, '-frames:v', '1'])
                self.engine = self.image_engine
                extension = '.png'
                if not self.context.request.format:
                    self.context.request.format = 'jpg'

        self.extension = extension
        self.engine.load(buffer, extension)

    def is_multiple(self):
        return False

    def cleanup(self):
        pass

    def __getattr__(self, attr):
        if not self.__dict__.get('engine'):
            raise AttributeError("'Engine' object has no attribute '%s'" % attr)
        return getattr(self.engine, attr)

    def __setattr__(self, attr, value):
        if attr in ('engine', 'ffmpeg_handle_animated_gif', 'use_gif_engine'):
            self.__dict__[attr] = value
        elif attr in ('context', 'extension'):
            self.__dict__[attr] = value
            if self.engine:
                setattr(self.engine, attr, value)
        elif self.engine:
            setattr(self.engine, attr, value)
        else:
            self.__dict__[attr] = value



from setuptools import setup, find_packages
from os import path

from io import open


CURR_DIR = path.abspath(path.dirname(__file__))


def get_long_description():
    with open(path.join(CURR_DIR, 'README.rst'), encoding='utf-8') as f:
        readme_rst = f.read()
        lines = readme_rst.split(u"\n")
        return u"\n".join(lines[10:])


setup(
    name='thumbor-video-engine',
    version='1.0.2',
    description='An engine and tools for manipulating videos with thumbor using ffmpeg',
    long_description=get_long_description(),
    long_description_content_type='text/x-rst',
    url='https://github.com/theatlantic/thumbor-video-engine',
    author='Frankie Dintino',
    author_email='fdintino@gmail.com',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Topic :: Software Development',
        'Environment :: Web Environment',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
    ],
    packages=find_packages('src'),
    package_dir={'': 'src'},
    install_requires=['thumbor'],
    python_requires='>=2.7, <3',
    zip_safe=False)
